# -*- coding: utf-8 -*-
import os
import sys
import unittest

from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive
from pydrive.test import test_util


class GoogleDriveFileListTest(unittest.TestCase):
  """Tests operations of files.GoogleDriveFileList class.
  Equivalent to Files.list in Google Drive API.
  """
  ga = GoogleAuth('settings/test1.yaml')
  ga.LocalWebserverAuth()
  drive = GoogleDrive(ga)

  def test_01_Files_List_GetList(self):
    drive = GoogleDrive(self.ga)
    flist = drive.ListFile({'q': "title = '%s' and trashed = false"
                                 % self.title})
    files = flist.GetList()  # Auto iterate every file
    for file1 in self.file_list:
      found = False
      for file2 in files:
        if file1['id'] == file2['id']:
          found = True
      self.assertEqual(found, True)

  def test_02_Files_List_ForLoop(self):
    drive = GoogleDrive(self.ga)
    flist = drive.ListFile({'q': "title = '%s' and trashed = false"%self.title,
                            'maxResults': 2})
    files = []
    for x in flist:  # Build iterator to access files simply with for loop
      self.assertTrue(len(x) <= 2)
      files.extend(x)
    for file1 in self.file_list:
      found = False
      for file2 in files:
        if file1['id'] == file2['id']:
          found = True
      self.assertEqual(found, True)

  def test_03_Files_List_GetList_Iterate(self):
    drive = GoogleDrive(self.ga)
    flist = drive.ListFile({'q': "title = '%s' and trashed = false"%self.title,
                            'maxResults': 2})
    files = []
    while True:
      try:
        x = flist.GetList()
        self.assertTrue(len(x) <= 2)
        files.extend(x)
      except StopIteration:
        break
    for file1 in self.file_list:
      found = False
      for file2 in files:
        if file1['id'] == file2['id']:
          found = True
      self.assertEqual(found, True)

  def test_File_List_Folders(self):
    drive = GoogleDrive(self.ga)
    folder1 = drive.CreateFile(
      {'mimeType': 'application/vnd.google-apps.folder',
       'title': self.title})
    folder1.Upload()
    self.file_list.append(folder1)

    flist = drive.ListFile({'q': "title = '%s' and trashed = false"
                                 % self.title})
    count = 0
    for _flist in flist:
      for file1 in _flist:
        self.assertFileInFileList(file1)
        count += 1

    self.assertTrue(count == 11)

  # setUp and tearDown methods.
  # ===========================
  def setUp(self):
    title = test_util.CreateRandomFileName()
    file_list = []
    for x in range(0, 10):
      file1 = self.drive.CreateFile()
      file1['title'] = title
      file1.Upload()
      file_list.append(file1)

    self.title = title
    self.file_list = file_list

  def tearDown(self):
    # Deleting uploaded files.
    for file1 in self.file_list:
      file1.Delete()

  def assertFileInFileList(self, file_object):
    found = False
    for file1 in self.file_list:
      if file_object['id'] == file1['id']:
        found = True
    self.assertEqual(found, True)

  def DeleteOldFile(self, file_name):
    try:
      os.remove(file_name)
    except OSError:
      pass

if __name__ == '__main__':
  unittest.main()



import unittest
import os
import time

from pydrive.auth import GoogleAuth

class GoogleAuthTest(unittest.TestCase):
  """Tests basic OAuth2 operations of auth.GoogleAuth."""

  def test_01_LocalWebserverAuthWithClientConfigFromFile(self):
    # Delete old credentials file
    self.DeleteOldCredentialsFile('credentials/1.dat')
    # Test if authentication works with config read from file
    ga = GoogleAuth('settings/test1.yaml')
    ga.LocalWebserverAuth()
    self.assertEqual(ga.access_token_expired, False)
    # Test if correct credentials file is created
    self.CheckCredentialsFile('credentials/1.dat')
    time.sleep(1)

  def test_02_LocalWebserverAuthWithClientConfigFromSettings(self):
    # Delete old credentials file
    self.DeleteOldCredentialsFile('credentials/2.dat')
    # Test if authentication works with config read from settings
    ga = GoogleAuth('settings/test2.yaml')
    ga.LocalWebserverAuth()
    self.assertEqual(ga.access_token_expired, False)
    # Test if correct credentials file is created
    self.CheckCredentialsFile('credentials/2.dat')
    time.sleep(1)

  def test_03_LocalWebServerAuthWithNoCredentialsSaving(self):
    # Delete old credentials file
    self.DeleteOldCredentialsFile('credentials/4.dat')
    # Provide wrong credentials file
    ga = GoogleAuth('settings/test3.yaml')
    ga.LocalWebserverAuth()
    self.assertEqual(ga.access_token_expired, False)
    # Test if correct credentials file is created
    self.CheckCredentialsFile('credentials/4.dat', no_file=True)
    time.sleep(1)

  def test_04_CommandLineAuthWithClientConfigFromFile(self):
    # Delete old credentials file
    self.DeleteOldCredentialsFile('credentials/1.dat')
    # Test if authentication works with config read from file
    ga = GoogleAuth('settings/test4.yaml')
    ga.CommandLineAuth()
    self.assertEqual(ga.access_token_expired, False)
    # Test if correct credentials file is created
    self.CheckCredentialsFile('credentials/1.dat')
    time.sleep(1)

  def test_05_ConfigFromSettingsWithoutOauthScope(self):
    # Test if authentication works without oauth_scope
    ga = GoogleAuth('settings/test5.yaml')
    ga.LocalWebserverAuth()
    self.assertEqual(ga.access_token_expired, False)
    time.sleep(1)

  def DeleteOldCredentialsFile(self, credentials):
    try:
      os.remove(credentials)
    except OSError:
      pass

  def CheckCredentialsFile(self, credentials, no_file=False):
    ga = GoogleAuth('settings/default.yaml')
    ga.LoadCredentialsFile(credentials)
    self.assertEqual(ga.access_token_expired, no_file)

if __name__ == '__main__':
  unittest.main()



from six import Iterator, iteritems

class ApiAttribute(object):
  """A data descriptor that sets and returns values."""

  def __init__(self, name):
    """Create an instance of ApiAttribute.

    :param name: name of this attribute.
    :type name: str.
    """
    self.name = name

  def __get__(self, obj, type=None):
    """Accesses value of this attribute."""
    return obj.attr.get(self.name)

  def __set__(self, obj, value):
    """Write value of this attribute."""
    obj.attr[self.name] = value
    if obj.dirty.get(self.name) is not None:
      obj.dirty[self.name] = True

  def __del__(self, obj=None):
    """Delete value of this attribute."""
    if not obj:
      return

    del obj.attr[self.name]
    if obj.dirty.get(self.name) is not None:
      del obj.dirty[self.name]

class ApiAttributeMixin(object):
  """Mixin to initialize required global variables to use ApiAttribute."""

  def __init__(self):
    self.attr = {}
    self.dirty = {}
    self.http = None  # Any element may make requests and will require this
    # field.


class ApiResource(dict):
  """Super class of all api resources.

  Inherits and behaves as a python dictionary to handle api resources.
  Save clean copy of metadata in self.metadata as a dictionary.
  Provides changed metadata elements to efficiently update api resources.
  """
  auth = ApiAttribute('auth')

  def __init__(self, *args, **kwargs):
    """Create an instance of ApiResource."""
    super(ApiResource, self).__init__()
    self.update(*args, **kwargs)
    self.metadata = dict(self)

  def __getitem__(self, key):
    """Overwritten method of dictionary.

    :param key: key of the query.
    :type key: str.
    :returns: value of the query.
    """
    return dict.__getitem__(self, key)

  def __setitem__(self, key, val):
    """Overwritten method of dictionary.

    :param key: key of the query.
    :type key: str.
    :param val: value of the query.
    """
    dict.__setitem__(self, key, val)

  def __repr__(self):
    """Overwritten method of dictionary."""
    dict_representation = dict.__repr__(self)
    return '%s(%s)' % (type(self).__name__, dict_representation)

  def update(self, *args, **kwargs):
    """Overwritten method of dictionary."""
    for k, v in iteritems(dict(*args, **kwargs)):
      self[k] = v

  def UpdateMetadata(self, metadata=None):
    """Update metadata and mark all of them to be clean."""
    if metadata:
      self.update(metadata)
    self.metadata = dict(self)

  def GetChanges(self):
    """Returns changed metadata elements to update api resources efficiently.

    :returns: dict -- changed metadata elements.
    """
    dirty = {}
    for key in self:
      if self.metadata.get(key) is None:
        dirty[key] = self[key]
      elif self.metadata[key] != self[key]:
        dirty[key] = self[key]
    return dirty


class ApiResourceList(ApiAttributeMixin, ApiResource, Iterator):
  """Abstract class of all api list resources.

  Inherits ApiResource and builds iterator to list any API resource.
  """
  metadata = ApiAttribute('metadata')

  def __init__(self, auth=None, metadata=None):
    """Create an instance of ApiResourceList.

    :param auth: authorized GoogleAuth instance.
    :type auth: GoogleAuth.
    :param metadata: parameter to send to list command.
    :type metadata: dict.
    """
    ApiAttributeMixin.__init__(self)
    ApiResource.__init__(self)
    self.auth = auth
    self.UpdateMetadata()
    if metadata:
      self.update(metadata)

  def __iter__(self):
    """Returns iterator object.

    :returns: ApiResourceList -- self
    """
    return self

  def __next__(self):
    """Make API call to list resources and return them.

    Auto updates 'pageToken' every time it makes API call and
    raises StopIteration when it reached the end of iteration.

    :returns: list -- list of API resources.
    :raises: StopIteration
    """
    if 'pageToken' in self and self['pageToken'] is None:
      raise StopIteration
    result = self._GetList()
    self['pageToken'] = self.metadata.get('nextPageToken')
    return result

  def GetList(self):
    """Get list of API resources.

    If 'maxResults' is not specified, it will automatically iterate through
    every resources available. Otherwise, it will make API call once and
    update 'pageToken'.

    :returns: list -- list of API resources.
    """
    if self.get('maxResults') is None:
      self['maxResults'] = 1000
      result = []
      for x in self:
        result.extend(x)
      del self['maxResults']
      return result
    else:
      return next(self)

  def _GetList(self):
    """Helper function which actually makes API call.

    Should be overwritten.

    :raises: NotImplementedError
    """
    raise NotImplementedError

  def Reset(self):
    """Resets current iteration"""
    if 'pageToken' in self:
      del self['pageToken']



# -*- coding: utf-8 -*-
import unittest

from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive


class GoogleDriveTest(unittest.TestCase):
    """Tests basic operations on meta-data information of the linked Google Drive account.
    """

    ga = GoogleAuth('settings/test1.yaml')
    ga.LocalWebserverAuth()

    def test_01_About_Request(self):
        drive = GoogleDrive(self.ga)

        about_object = drive.GetAbout()
        self.assertTrue(about_object is not None, "About object not loading.")


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import filecmp
import os
import unittest
from io import BytesIO

from six.moves import range
import timeout_decorator
from concurrent.futures import ThreadPoolExecutor, as_completed

from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive
from pydrive.files import ApiRequestError, GoogleDriveFile

import test_util


class GoogleDriveFileTest(unittest.TestCase):
  """Tests basic file operations of files.GoogleDriveFile.
  Upload and download of contents and metadata, and thread-safety checks.
  Equivalent to Files.insert, Files.update, Files.patch in Google Drive API.
  """

  ga = GoogleAuth('settings/test1.yaml')
  ga.LocalWebserverAuth()
  first_file = 'a.png'
  second_file = 'b.png'

  def test_01_Files_Insert(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'firsttestfile'
    file1['title'] = filename
    file1.Upload()  # Files.insert

    self.assertEqual(file1.metadata['title'], filename)
    file2 = drive.CreateFile({'id': file1['id']})  # Download file from id.
    self.assertEqual(file2['title'], filename)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_02_Files_Insert_Unicode(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = u'첫번째 파일'
    file1['title'] = filename
    file1.Upload()  # Files.insert

    self.assertEqual(file1.metadata['title'], filename)
    file2 = drive.CreateFile({'id': file1['id']})  # Download file from id.
    self.assertEqual(file2['title'], filename)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_03_Files_Insert_Content_String(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'secondtestfile'
    content = 'hello world!'
    file1['title'] = filename
    file1.SetContentString(content)
    file1.Upload()  # Files.insert

    self.assertEqual(file1.GetContentString(), content)

    file1.FetchContent()  # Force download and double check content
    self.assertEqual(file1.metadata['title'], filename)
    self.assertEqual(file1.GetContentString(), content)

    file2 = drive.CreateFile({'id': file1['id']})  # Download file from id.
    self.assertEqual(file2.GetContentString(), content)
    self.assertEqual(file2.metadata['title'], filename)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_04_Files_Insert_Content_Unicode_String(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = u'두번째 파일'
    content = u'안녕 세상아!'
    file1['title'] = filename
    file1.SetContentString(content)
    file1.Upload()  # Files.insert

    self.assertEqual(file1.GetContentString(), content)
    self.assertEqual(file1.metadata['title'], filename)
    file1.FetchContent()  # Force download and double check content.
    self.assertEqual(file1.GetContentString(), content)

    file2 = drive.CreateFile({'id': file1['id']})  # Download file from id.
    self.assertEqual(file2.GetContentString(), content)
    self.assertEqual(file2.metadata['title'], filename)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_05_Files_Insert_Content_File(self):
    self.DeleteOldFile(self.first_file+'1')
    self.DeleteOldFile(self.first_file+'2')
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'filecontent'
    file1['title'] = filename
    file1.SetContentFile(self.first_file)
    file1.Upload()  # Files.insert

    self.assertEqual(file1.metadata['title'], filename)
    file1.FetchContent()  # Force download and double check content.
    file1.GetContentFile(self.first_file+'1')
    self.assertEqual(filecmp.cmp(self.first_file, self.first_file+'1'), True)

    file2 = drive.CreateFile({'id': file1['id']})  # Download file from id.
    file2.GetContentFile(self.first_file+'2')
    self.assertEqual(filecmp.cmp(self.first_file, self.first_file+'2'), True)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_06_Files_Patch(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'prepatchtestfile'
    newfilename = 'patchtestfile'
    file1['title'] = filename
    file1.Upload()  # Files.insert

    self.assertEqual(file1.metadata['title'], filename)
    file1['title'] = newfilename
    file1.Upload()  # Files.patch

    self.assertEqual(file1.metadata['title'], newfilename)
    file2 = drive.CreateFile({'id': file1['id']})  # Download file from id.
    file2.FetchMetadata()
    self.assertEqual(file2.metadata['title'], newfilename)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_07_Files_Patch_Skipping_Content(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'prepatchtestfile'
    newfilename = 'patchtestfile'
    content = 'hello world!'

    file1['title'] = filename
    file1.SetContentString(content)
    file1.Upload()  # Files.insert
    self.assertEqual(file1.metadata['title'], filename)

    file1['title'] = newfilename
    file1.Upload()  # Files.patch
    self.assertEqual(file1.metadata['title'], newfilename)
    self.assertEqual(file1.GetContentString(), content)
    self.assertEqual(file1.GetContentString(), content)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_08_Files_Update_String(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'preupdatetestfile'
    newfilename = 'updatetestfile'
    content = 'hello world!'
    newcontent = 'hello new world!'

    file1['title'] = filename
    file1.SetContentString(content)
    file1.Upload()  # Files.insert
    self.assertEqual(file1.metadata['title'], filename)
    self.assertEqual(file1.GetContentString(), content)

    file1.FetchContent()  # Force download and double check content.
    self.assertEqual(file1.GetContentString(), content)

    file1['title'] = newfilename
    file1.SetContentString(newcontent)
    file1.Upload()  # Files.update
    self.assertEqual(file1.metadata['title'], newfilename)
    self.assertEqual(file1.GetContentString(), newcontent)
    self.assertEqual(file1.GetContentString(), newcontent)

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_09_Files_Update_File(self):
    self.DeleteOldFile(self.first_file+'1')
    self.DeleteOldFile(self.second_file+'1')
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    filename = 'preupdatetestfile'
    newfilename = 'updatetestfile'

    file1['title'] = filename
    file1.SetContentFile(self.first_file)
    file1.Upload()  # Files.insert
    self.assertEqual(file1.metadata['title'], filename)

    file1.FetchContent()  # Force download and double check content.
    file1.GetContentFile(self.first_file+'1')
    self.assertEqual(filecmp.cmp(self.first_file, self.first_file+'1'), True)

    file1['title'] = newfilename
    file1.SetContentFile(self.second_file)
    file1.Upload()  # Files.update
    self.assertEqual(file1.metadata['title'], newfilename)
    file1.GetContentFile(self.second_file+'1')
    self.assertEqual(filecmp.cmp(self.second_file, self.second_file+'1'), True)

    self.DeleteUploadedFiles(drive, [file1['id']])

  # Tests for Trash/UnTrash/Delete.
  # ===============================

  def test_Files_Trash_File(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    self.assertFalse(file1.metadata[u'labels'][u'trashed'])

    # Download to verify non-trashed state on GDrive.
    file2 = drive.CreateFile({'id': file1['id']})
    file2.FetchMetadata()
    self.assertFalse(file2.metadata[u'labels'][u'trashed'])

    file1.Trash()
    self.assertTrue(file1.metadata[u'labels'][u'trashed'])

    file2.FetchMetadata()
    self.assertTrue(file2.metadata[u'labels'][u'trashed'])

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_Files_Trash_File_Just_ID(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    self.assertFalse(file1.metadata[u'labels'][u'trashed'])

    # Trash file by ID.
    file2 = drive.CreateFile({'id': file1['id']})
    file2.Trash()

    # Verify trashed by downloading metadata.
    file1.FetchMetadata()
    self.assertTrue(file1.metadata[u'labels'][u'trashed'])

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_Files_UnTrash_File(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    file1.Trash()
    self.assertTrue(file1.metadata[u'labels'][u'trashed'])

    # Verify that file is trashed by downloading metadata.
    file2 = drive.CreateFile({'id': file1['id']})
    file2.FetchMetadata()
    self.assertTrue(file2.metadata[u'labels'][u'trashed'])

    # Un-trash the file, and assert local metadata is updated correctly.
    file1.UnTrash()
    self.assertFalse(file1.metadata[u'labels'][u'trashed'])

    # Re-fetch the metadata, and assert file un-trashed on GDrive.
    file2.FetchMetadata()
    self.assertFalse(file2.metadata[u'labels'][u'trashed'])

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_Files_UnTrash_File_Just_ID(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    file1.Trash()
    self.assertTrue(file1.metadata[u'labels'][u'trashed'])

    file2 = drive.CreateFile({'id': file1['id']})
    file2.UnTrash()  # UnTrash without fetching metadata.

    file1.FetchMetadata()
    self.assertFalse(file1.metadata[u'labels'][u'trashed'])

    self.DeleteUploadedFiles(drive, [file1['id']])

  def test_Files_Delete_File(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    file2 = drive.CreateFile({'id': file1['id']})

    file1.Delete()

    try:
      file2.FetchMetadata()
      self.fail("File not deleted correctly.")
    except ApiRequestError as e:
      pass

  def test_Files_Delete_File_Just_ID(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    file2 = drive.CreateFile({'id': file1['id']})

    file2.Delete()

    try:
      file1.FetchMetadata()
      self.fail("File not deleted correctly.")
    except ApiRequestError as e:
      pass

  # Tests for Permissions.
  # ======================

  def test_Files_FetchMetadata_Fields(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()

    self.assertFalse('permissions' in file1)

    file1.FetchMetadata('permissions')
    self.assertTrue('permissions' in file1)
    file1.Delete()

  def test_Files_Insert_Permission(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()

    # Verify only one permission before inserting permission.
    permissions = file1.GetPermissions()
    self.assertEqual(len(permissions), 1)
    self.assertEqual(len(file1['permissions']), 1)

    # Insert the permission.
    permission = file1.InsertPermission({'type': 'anyone',
                            'value': 'anyone',
                            'role': 'reader'})
    self.assertTrue(permission)
    self.assertEqual(len(file1["permissions"]), 2)
    self.assertEqual(file1["permissions"][1]["type"], "anyone")

    permissions = file1.GetPermissions()
    self.assertEqual(len(file1["permissions"]), 2)
    self.assertEqual(file1["permissions"][1]["type"], "anyone")
    self.assertEqual(permissions[1]["type"], "anyone")

    # Verify remote changes made.
    file2 = drive.CreateFile({'id': file1['id']})
    permissions = file2.GetPermissions()
    self.assertEqual(len(permissions), 2)
    self.assertEqual(permissions[1]["type"], "anyone")

    file1.Delete()

  def test_Files_Get_Permissions(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()

    self.assertFalse('permissions' in file1)

    permissions = file1.GetPermissions()
    self.assertTrue(permissions is not None)
    self.assertTrue('permissions' in file1)

    file1.Delete()

  def test_Files_Delete_Permission(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()
    file1.InsertPermission({'type': 'anyone',
                            'value': 'anyone',
                            'role': 'reader'})
    permissions = file1.GetPermissions()
    self.assertEqual(len(permissions), 2)
    self.assertEqual(len(file1['permissions']), 2)

    file1.DeletePermission(permissions[1]['id'])
    self.assertEqual(len(file1['permissions']), 1)

    # Verify remote changes made.
    file2 = drive.CreateFile({'id': file1['id']})
    permissions = file2.GetPermissions()
    self.assertEqual(len(permissions), 1)

    file1.Delete()

  def test_Files_Delete_Permission_Invalid(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()
    file1.Upload()

    try:
      file1.DeletePermission('invalid id')
      self.fail("Deleting invalid permission not raising exception.")
    except ApiRequestError as e:
      pass

    file1.Delete()

  def test_GFile_Conversion_Lossless_String(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()

    # Upload a string, and convert into Google Doc format.
    test_string = 'Generic, non-exhaustive ASCII test string.'
    file1.SetContentString(test_string)
    file1.Upload({'convert': True})

    # Download string as plain text.
    downloaded_string = file1.GetContentString(mimetype='text/plain')
    self.assertEqual(test_string, downloaded_string, "Strings do not match")

    # Download content into file and ensure that file content matches original
    # content string.
    downloaded_file_name = '_tmp_downloaded_file_name.txt'
    file1.GetContentFile(downloaded_file_name, mimetype='text/plain')
    downloaded_string = open(downloaded_file_name).read()
    self.assertEqual(test_string, downloaded_string, "Strings do not match")

    # Delete temp file.
    os.remove(downloaded_file_name)

  # Tests for GDrive conversion.
  # ============================

  def setup_gfile_conversion_test(self):
    drive = GoogleDrive(self.ga)
    file1 = drive.CreateFile()

    # Create a file to upload.
    file_name = '_tmp_source_file.txt'
    downloaded_file_name = '_tmp_downloaded_file_name.txt'
    original_file_content = 'Generic, non-exhaustive\n ASCII test string.'
    source_file = open(file_name, mode='w+')
    source_file.write(original_file_content)
    source_file.close()
    original_file_content = test_util.StripNewlines(original_file_content)

    return file1, file_name, original_file_content, downloaded_file_name

  def cleanup_gfile_conversion_test(self, file1, file_name,
                                    downloaded_file_name):
    # Delete temporary files.
    os.path.exists(file_name) and os.remove(file_name)
    os.path.exists(downloaded_file_name) and os.remove(downloaded_file_name)
    file1.Delete()  # Delete uploaded file.

  def test_GFile_Conversion_Remove_BOM(self):
    (file1, file_name, original_file_content, downloaded_file_name) = \
        self.setup_gfile_conversion_test()
    try:
      # Upload source_file and convert into Google Doc format.
      file1.SetContentFile(file_name)
      file1.Upload({'convert': True})

      # Download as string.
      downloaded_content_no_bom = file1.GetContentString(mimetype='text/plain',
                                                         remove_bom=True)
      downloaded_content_no_bom = test_util.StripNewlines(
        downloaded_content_no_bom)
      self.assertEqual(original_file_content, downloaded_content_no_bom)

      # Download as file.
      file1.GetContentFile(downloaded_file_name, remove_bom=True)
      downloaded_content = open(downloaded_file_name).read()
      downloaded_content = test_util.StripNewlines(downloaded_content)
      self.assertEqual(original_file_content, downloaded_content)

    finally:
      self.cleanup_gfile_conversion_test(file1, file_name, downloaded_file_name)

  def test_Gfile_Conversion_Add_Remove_BOM(self):
    """Tests whether you can switch between the BOM appended and removed
    version on the fly."""
    (file1, file_name, original_file_content, downloaded_file_name) = \
        self.setup_gfile_conversion_test()
    try:
      file1.SetContentFile(file_name)
      file1.Upload({'convert': True})

      content_bom = file1.GetContentString(mimetype='text/plain')
      content_no_bom = file1.GetContentString(mimetype='text/plain',
                                              remove_bom=True)
      content_bom_2 = file1.GetContentString(mimetype='text/plain')

      self.assertEqual(content_bom, content_bom_2)
      self.assertNotEqual(content_bom, content_no_bom)
      self.assertTrue(len(content_bom) > len(content_no_bom))

    finally:
      self.cleanup_gfile_conversion_test(file1, file_name, downloaded_file_name)

  def test_InsertPrefix(self):
    # Create BytesIO.
    file_obj = BytesIO('abc')
    original_length = len(file_obj.getvalue())
    char_to_insert = u'\ufeff'.encode('utf8')

    # Insert the prefix.
    GoogleDriveFile._InsertPrefix(file_obj, char_to_insert)
    modified_length = len(file_obj.getvalue())
    self.assertGreater(modified_length, original_length)
    self.assertEqual(file_obj.getvalue(), u'\ufeffabc'.encode('utf8'))

  def test_InsertPrefixLarge(self):
    # Create BytesIO.
    test_content = 'abc' * 800
    file_obj = BytesIO(test_content)
    original_length = len(file_obj.getvalue())
    char_to_insert = u'\ufeff'.encode('utf8')

    # Insert the prefix.
    GoogleDriveFile._InsertPrefix(file_obj, char_to_insert)
    modified_length = len(file_obj.getvalue())
    self.assertGreater(modified_length, original_length)
    expected_content = u'\ufeff' + test_content
    self.assertEqual(file_obj.getvalue(), expected_content.encode('utf8'))

  def test_RemovePrefix(self):
    # Create BytesIO.
    file_obj = BytesIO(u'\ufeffabc'.encode('utf8'))
    original_length = len(file_obj.getvalue())
    char_to_remove = u'\ufeff'.encode('utf8')

    # Insert the prefix.
    GoogleDriveFile._RemovePrefix(file_obj, char_to_remove)
    modified_length = len(file_obj.getvalue())
    self.assertLess(modified_length, original_length)
    self.assertEqual(file_obj.getvalue(), 'abc')

  def test_RemovePrefixLarge(self):
    # Create BytesIO.
    test_content = u'\ufeff' + u'abc' * 800
    file_obj = BytesIO(test_content.encode('utf8'))
    original_length = len(file_obj.getvalue())
    char_to_remove = u'\ufeff'.encode('utf8')

    # Insert the prefix.
    GoogleDriveFile._RemovePrefix(file_obj, char_to_remove)
    modified_length = len(file_obj.getvalue())
    self.assertLess(modified_length, original_length)
    self.assertEqual(file_obj.getvalue(), test_content[1:])

  # Setup for concurrent upload testing.
  # =====================================
  class UploadWorker:
    def __init__(self, gdrive_file, generate_http=False):
      self.gdrive_file = gdrive_file
      self.param = {}
      if generate_http:
        self.param = {"http": gdrive_file.auth.Get_Http_Object()}

    def run(self):
      self.gdrive_file.Upload(param=self.param)

  def _parallel_uploader(self, num_of_uploads, num_of_workers,
                         use_per_thread_http=False):
    drive = GoogleDrive(self.ga)
    thread_pool = ThreadPoolExecutor(max_workers=num_of_workers)

    # Create list of gdrive_files.
    upload_files = []
    remote_name = test_util.CreateRandomFileName()
    for i in range(num_of_uploads):
      file_name = self.first_file if i % 2 == 0 else self.second_file
      up_file = drive.CreateFile()
      up_file['title'] = remote_name
      up_file.SetContentFile(file_name)
      upload_files.append(up_file)

    # Ensure there are no files with the random file name.
    files = drive.ListFile(param={'q': "title = '%s' and trashed = false" %
                               remote_name}).GetList()
    self.assertTrue(len(files) == 0)

    # Submit upload jobs to ThreadPoolExecutor.
    futures = []
    for i in range(num_of_uploads):
      upload_worker = self.UploadWorker(upload_files[i],
                                        use_per_thread_http)
      futures.append(thread_pool.submit(upload_worker.run))

    # Ensure that all threads a) return, and b) encountered no exceptions.
    for future in as_completed(futures):
      self.assertIsNone(future.exception())
    thread_pool.shutdown()

    # Ensure 10 files were uploaded.
    files = drive.ListFile(param={'q': "title = '%s' and trashed = false" %
                                       remote_name}).GetList()
    self.assertTrue(len(files) == 10)

    # Remove uploaded files.
    self.DeleteUploadedFiles(drive, [fi['id'] for fi in upload_files])

  @timeout_decorator.timeout(80)
  def test_Parallel_Files_Insert_File_Auto_Generated_HTTP(self):
    self._parallel_uploader(10, 10)

  @timeout_decorator.timeout(80)
  def test_Parallel_Insert_File_Passed_HTTP(self):
    self._parallel_uploader(10, 10, True)

  # Helper functions.
  # =================

  def DeleteOldFile(self, file_name):
    try:
      os.remove(file_name)
    except OSError:
      pass

  def DeleteUploadedFiles(self, drive, ids):
    for element in ids:
      tmp_file = drive.CreateFile({'id': element})
      tmp_file.Delete()


if __name__ == '__main__':
  unittest.main()



from .apiattr import ApiAttributeMixin
from .files import GoogleDriveFile
from .files import GoogleDriveFileList
from .auth import LoadAuth


class GoogleDrive(ApiAttributeMixin, object):
  """Main Google Drive class."""

  def __init__(self, auth=None):
    """Create an instance of GoogleDrive.

    :param auth: authorized GoogleAuth instance.
    :type auth: pydrive.auth.GoogleAuth.
    """
    ApiAttributeMixin.__init__(self)
    self.auth = auth

  def CreateFile(self, metadata=None):
    """Create an instance of GoogleDriveFile with auth of this instance.

    This method would not upload a file to GoogleDrive.

    :param metadata: file resource to initialize GoogleDriveFile with.
    :type metadata: dict.
    :returns: pydrive.files.GoogleDriveFile -- initialized with auth of this instance.
    """
    return GoogleDriveFile(auth=self.auth, metadata=metadata)

  def ListFile(self, param=None):
    """Create an instance of GoogleDriveFileList with auth of this instance.

    This method will not fetch from Files.List().

    :param param: parameter to be sent to Files.List().
    :type param: dict.
    :returns: pydrive.files.GoogleDriveFileList -- initialized with auth of this instance.
    """
    return GoogleDriveFileList(auth=self.auth, param=param)

  @LoadAuth
  def GetAbout(self):
    """Return information about the Google Drive of the auth instance.

    :returns: A dictionary of Google Drive information like user, usage, quota etc.
    """
    return self.auth.service.about().get().execute(http=self.http)



import socket
import webbrowser
import httplib2
import oauth2client.clientsecrets as clientsecrets
from six.moves import input

from apiclient.discovery import build
from functools import wraps
from oauth2client.service_account import ServiceAccountCredentials
from oauth2client.client import FlowExchangeError
from oauth2client.client import AccessTokenRefreshError
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.client import OOB_CALLBACK_URN
from oauth2client.file import Storage
from oauth2client.tools import ClientRedirectHandler
from oauth2client.tools import ClientRedirectServer
from oauth2client._helpers import scopes_to_string
from .apiattr import ApiAttribute
from .apiattr import ApiAttributeMixin
from .settings import LoadSettingsFile
from .settings import ValidateSettings
from .settings import SettingsError
from .settings import InvalidConfigError


class AuthError(Exception):
  """Base error for authentication/authorization errors."""


class InvalidCredentialsError(IOError):
  """Error trying to read credentials file."""


class AuthenticationRejected(AuthError):
  """User rejected authentication."""


class AuthenticationError(AuthError):
  """General authentication error."""


class RefreshError(AuthError):
  """Access token refresh error."""

def LoadAuth(decoratee):
  """Decorator to check if the auth is valid and loads auth if not."""
  @wraps(decoratee)
  def _decorated(self, *args, **kwargs):
    # Initialize auth if needed.
    if self.auth is None:
      self.auth = GoogleAuth()
    # Re-create access token if it expired.
    if self.auth.access_token_expired:
      if getattr(self, 'auth_method', False) == 'service':
        self.auth.ServiceAuth()
      else:
        self.auth.LocalWebserverAuth()

    # Initialise service if not built yet.
    if self.auth.service is None:
      self.auth.Authorize()

    # Ensure that a thread-safe HTTP object is provided.
    if kwargs is not None and \
            "param" in kwargs and \
            kwargs["param"] is not None and \
            "http" in kwargs["param"] and \
            kwargs["param"]["http"] is not None:
      self.http = kwargs["param"]["http"]
      del kwargs["param"]["http"]

    else:  # If HTTP object not specified, each call creates new HTTP object.
      self.http = self.auth.Get_Http_Object()

    return decoratee(self, *args, **kwargs)
  return _decorated

def CheckServiceAuth(decoratee):
  """Decorator to authorize service account."""
  @wraps(decoratee)
  def _decorated(self, *args, **kwargs):
    self.auth_method = 'service'
    dirty = False
    save_credentials = self.settings.get('save_credentials')
    if self.credentials is None and save_credentials:
      self.LoadCredentials()
    if self.credentials is None:
      decoratee(self, *args, **kwargs)
      self.Authorize()
      dirty = True
    else:
      if self.access_token_expired:
        if self.credentials.refresh_token is not None:
          self.Refresh()
        else:
          decoratee(self, *args, **kwargs)
          self.Authorize()
        dirty = True
    if dirty and save_credentials:
      self.SaveCredentials()
  return _decorated

def CheckAuth(decoratee):
  """Decorator to check if it requires OAuth2 flow request."""
  @wraps(decoratee)
  def _decorated(self, *args, **kwargs):
    dirty = False
    code = None
    save_credentials = self.settings.get('save_credentials')
    if self.credentials is None and save_credentials:
      self.LoadCredentials()
    if self.flow is None:
      self.GetFlow()
    if self.credentials is None:
      code = decoratee(self, *args, **kwargs)
      dirty = True
    else:
      if self.access_token_expired:
        if self.credentials.refresh_token is not None:
          self.Refresh()
        else:
          code = decoratee(self, *args, **kwargs)
        dirty = True
    if code is not None:
      self.Auth(code)
    if dirty and save_credentials:
      self.SaveCredentials()
  return _decorated


class GoogleAuth(ApiAttributeMixin, object):
  """Wrapper class for oauth2client library in google-api-python-client.

  Loads all settings and credentials from one 'settings.yaml' file
  and performs common OAuth2.0 related functionality such as authentication
  and authorization.
  """
  DEFAULT_SETTINGS = {
      'client_config_backend': 'file',
      'client_config_file': 'client_secrets.json',
      'save_credentials': False,
      'oauth_scope': ['https://www.googleapis.com/auth/drive']
      }
  CLIENT_CONFIGS_LIST = ['client_id', 'client_secret', 'auth_uri',
                         'token_uri', 'revoke_uri', 'redirect_uri']
  SERVICE_CONFIGS_LIST = ['client_service_email', 'client_user_email',
                          'client_pkcs12_file_path']
  settings = ApiAttribute('settings')
  client_config = ApiAttribute('client_config')
  flow = ApiAttribute('flow')
  credentials = ApiAttribute('credentials')
  http = ApiAttribute('http')
  service = ApiAttribute('service')
  auth_method = ApiAttribute('auth_method')

  def __init__(self, settings_file='settings.yaml',http_timeout=None):
    """Create an instance of GoogleAuth.

    This constructor just sets the path of settings file.
    It does not actually read the file.

    :param settings_file: path of settings file. 'settings.yaml' by default.
    :type settings_file: str.
    """
    self.http_timeout=http_timeout
    ApiAttributeMixin.__init__(self)
    self.client_config = {}
    try:
      self.settings = LoadSettingsFile(settings_file)
    except SettingsError:
      self.settings = self.DEFAULT_SETTINGS
    else:
      if self.settings is None:
        self.settings = self.DEFAULT_SETTINGS
      else:
        ValidateSettings(self.settings)

  @property
  def access_token_expired(self):
    """Checks if access token doesn't exist or is expired.

    :returns: bool -- True if access token doesn't exist or is expired.
    """
    if self.credentials is None:
      return True
    return self.credentials.access_token_expired

  @CheckAuth
  def LocalWebserverAuth(self, host_name='localhost',
                         port_numbers=None):
    """Authenticate and authorize from user by creating local web server and
    retrieving authentication code.

    This function is not for web server application. It creates local web server
    for user from standalone application.

    :param host_name: host name of the local web server.
    :type host_name: str.
    :param port_numbers: list of port numbers to be tried to used.
    :type port_numbers: list.
    :returns: str -- code returned from local web server
    :raises: AuthenticationRejected, AuthenticationError
    """
    if port_numbers is None:
      port_numbers = [8080, 8090]  # Mutable objects should not be default
      # values, as each call's changes are global.
    success = False
    port_number = 0
    for port in port_numbers:
      port_number = port
      try:
        httpd = ClientRedirectServer((host_name, port), ClientRedirectHandler)
      except socket.error as e:
        pass
      else:
        success = True
        break
    if success:
      oauth_callback = 'http://%s:%s/' % (host_name, port_number)
    else:
      print('Failed to start a local web server. Please check your firewall')
      print('settings and locally running programs that may be blocking or')
      print('using configured ports. Default ports are 8080 and 8090.')
      raise AuthenticationError()
    self.flow.redirect_uri = oauth_callback
    authorize_url = self.GetAuthUrl()
    webbrowser.open(authorize_url, new=1, autoraise=True)
    print('Your browser has been opened to visit:')
    print()
    print('    ' + authorize_url)
    print()
    httpd.handle_request()
    if 'error' in httpd.query_params:
      print('Authentication request was rejected')
      raise AuthenticationRejected('User rejected authentication')
    if 'code' in httpd.query_params:
      return httpd.query_params['code']
    else:
      print('Failed to find "code" in the query parameters of the redirect.')
      print('Try command-line authentication')
      raise AuthenticationError('No code found in redirect')

  @CheckAuth
  def CommandLineAuth(self):
    """Authenticate and authorize from user by printing authentication url
    retrieving authentication code from command-line.

    :returns: str -- code returned from commandline.
    """
    self.flow.redirect_uri = OOB_CALLBACK_URN
    authorize_url = self.GetAuthUrl()
    print('Go to the following link in your browser:')
    print()
    print('    ' + authorize_url)
    print()
    return input('Enter verification code: ').strip()

  @CheckServiceAuth
  def ServiceAuth(self):
    """Authenticate and authorize using P12 private key, client id
    and client email for a Service account.
    :raises: AuthError, InvalidConfigError
    """
    if set(self.SERVICE_CONFIGS_LIST) - set(self.client_config):
      self.LoadServiceConfigSettings()
    scopes = scopes_to_string(self.settings['oauth_scope'])
    user_email = self.client_config.get('client_user_email')
    service_email = self.client_config['client_service_email']
    file_path = self.client_config['client_pkcs12_file_path']
    self.credentials = ServiceAccountCredentials.from_p12_keyfile(
        service_account_email=service_email,
        filename=file_path,
        scopes=scopes)
    if user_email:
        self.credentials = self.credentials.create_delegated(sub=user_email)

  def LoadCredentials(self, backend=None):
    """Loads credentials or create empty credentials if it doesn't exist.

    :param backend: target backend to save credential to.
    :type backend: str.
    :raises: InvalidConfigError
    """
    if backend is None:
      backend = self.settings.get('save_credentials_backend')
      if backend is None:
        raise InvalidConfigError('Please specify credential backend')
    if backend == 'file':
      self.LoadCredentialsFile()
    else:
      raise InvalidConfigError('Unknown save_credentials_backend')

  def LoadCredentialsFile(self, credentials_file=None):
    """Loads credentials or create empty credentials if it doesn't exist.

    Loads credentials file from path in settings if not specified.

    :param credentials_file: path of credentials file to read.
    :type credentials_file: str.
    :raises: InvalidConfigError, InvalidCredentialsError
    """
    if credentials_file is None:
      credentials_file = self.settings.get('save_credentials_file')
      if credentials_file is None:
        raise InvalidConfigError('Please specify credentials file to read')
    try:
      storage = Storage(credentials_file)
      self.credentials = storage.get()
    except IOError:
      raise InvalidCredentialsError('Credentials file cannot be symbolic link')

  def SaveCredentials(self, backend=None):
    """Saves credentials according to specified backend.

    If you have any specific credentials backend in mind, don't use this
    function and use the corresponding function you want.

    :param backend: backend to save credentials.
    :type backend: str.
    :raises: InvalidConfigError
    """
    if backend is None:
      backend = self.settings.get('save_credentials_backend')
      if backend is None:
        raise InvalidConfigError('Please specify credential backend')
    if backend == 'file':
      self.SaveCredentialsFile()
    else:
      raise InvalidConfigError('Unknown save_credentials_backend')

  def SaveCredentialsFile(self, credentials_file=None):
    """Saves credentials to the file in JSON format.

    :param credentials_file: destination to save file to.
    :type credentials_file: str.
    :raises: InvalidConfigError, InvalidCredentialsError
    """
    if self.credentials is None:
      raise InvalidCredentialsError('No credentials to save')
    if credentials_file is None:
      credentials_file = self.settings.get('save_credentials_file')
      if credentials_file is None:
        raise InvalidConfigError('Please specify credentials file to read')
    try:
      storage = Storage(credentials_file)
      storage.put(self.credentials)
      self.credentials.set_store(storage)
    except IOError:
      raise InvalidCredentialsError('Credentials file cannot be symbolic link')

  def LoadClientConfig(self, backend=None):
    """Loads client configuration according to specified backend.

    If you have any specific backend to load client configuration from in mind,
    don't use this function and use the corresponding function you want.

    :param backend: backend to load client configuration from.
    :type backend: str.
    :raises: InvalidConfigError
    """
    if backend is None:
      backend = self.settings.get('client_config_backend')
      if backend is None:
        raise InvalidConfigError('Please specify client config backend')
    if backend == 'file':
      self.LoadClientConfigFile()
    elif backend == 'settings':
      self.LoadClientConfigSettings()
    elif backend == 'service':
      self.LoadServiceConfigSettings()
    else:
      raise InvalidConfigError('Unknown client_config_backend')

  def LoadClientConfigFile(self, client_config_file=None):
    """Loads client configuration file downloaded from APIs console.

    Loads client config file from path in settings if not specified.

    :param client_config_file: path of client config file to read.
    :type client_config_file: str.
    :raises: InvalidConfigError
    """
    if client_config_file is None:
      client_config_file = self.settings['client_config_file']
    try:
      client_type, client_info = clientsecrets.loadfile(client_config_file)
    except clientsecrets.InvalidClientSecretsError as error:
      raise InvalidConfigError('Invalid client secrets file %s' % error)
    if not client_type in (clientsecrets.TYPE_WEB,
                           clientsecrets.TYPE_INSTALLED):
      raise InvalidConfigError('Unknown client_type of client config file')

    # General settings.
    try:
      config_index = ['client_id', 'client_secret', 'auth_uri', 'token_uri']
      for config in config_index:
        self.client_config[config] = client_info[config]

      self.client_config['revoke_uri'] = client_info.get('revoke_uri')
      self.client_config['redirect_uri'] = client_info['redirect_uris'][0]
    except KeyError:
      raise InvalidConfigError('Insufficient client config in file')

    # Service auth related fields.
    service_auth_config = ['client_email']
    try:
      for config in service_auth_config:
        self.client_config[config] = client_info[config]
    except KeyError:
      pass  # The service auth fields are not present, handling code can go here.

  def LoadServiceConfigSettings(self):
    """Loads client configuration from settings file.
    :raises: InvalidConfigError
    """
    for config in self.SERVICE_CONFIGS_LIST:
      try:
        self.client_config[config] = \
          self.settings['service_config'][config]
      except KeyError:
        err = "Insufficient service config in settings"
        err += "\n\nMissing: {} key.".format(config)
        raise InvalidConfigError(err)

  def LoadClientConfigSettings(self):
    """Loads client configuration from settings file.

    :raises: InvalidConfigError
    """
    for config in self.CLIENT_CONFIGS_LIST:
      try:
        self.client_config[config] = self.settings['client_config'][config]
      except KeyError:
        raise InvalidConfigError('Insufficient client config in settings')

  def GetFlow(self):
    """Gets Flow object from client configuration.

    :raises: InvalidConfigError
    """
    if not all(config in self.client_config \
               for config in self.CLIENT_CONFIGS_LIST):
      self.LoadClientConfig()
    constructor_kwargs = {
        'redirect_uri': self.client_config['redirect_uri'],
        'auth_uri': self.client_config['auth_uri'],
        'token_uri': self.client_config['token_uri'],
    }
    if self.client_config['revoke_uri'] is not None:
      constructor_kwargs['revoke_uri'] = self.client_config['revoke_uri']
    self.flow = OAuth2WebServerFlow(
        self.client_config['client_id'],
        self.client_config['client_secret'],
        scopes_to_string(self.settings['oauth_scope']),
        **constructor_kwargs)
    if self.settings.get('get_refresh_token'):
      self.flow.params.update({
          'access_type': 'offline',
          'approval_prompt': 'force'
      })

  def Refresh(self):
    """Refreshes the access_token.

    :raises: RefreshError
    """
    if self.credentials is None:
      raise RefreshError('No credential to refresh.')
    if self.credentials.refresh_token is None:
      raise RefreshError('No refresh_token found.'
                         'Please set access_type of OAuth to offline.')
    if self.http is None:
      self.http = httplib2.Http(timeout=self.http_timeout)
    try:
      self.credentials.refresh(self.http)
    except AccessTokenRefreshError as error:
      raise RefreshError('Access token refresh failed: %s' % error)

  def GetAuthUrl(self):
    """Creates authentication url where user visits to grant access.

    :returns: str -- Authentication url.
    """
    if self.flow is None:
      self.GetFlow()
    return self.flow.step1_get_authorize_url()

  def Auth(self, code):
    """Authenticate, authorize, and build service.

    :param code: Code for authentication.
    :type code: str.
    :raises: AuthenticationError
    """
    self.Authenticate(code)
    self.Authorize()

  def Authenticate(self, code):
    """Authenticates given authentication code back from user.

    :param code: Code for authentication.
    :type code: str.
    :raises: AuthenticationError
    """
    if self.flow is None:
      self.GetFlow()
    try:
      self.credentials = self.flow.step2_exchange(code)
    except FlowExchangeError as e:
      raise AuthenticationError('OAuth2 code exchange failed: %s' % e)
    print('Authentication successful.')

  def Authorize(self):
    """Authorizes and builds service.

    :raises: AuthenticationError
    """
    if self.http is None:
      self.http = httplib2.Http(timeout=self.http_timeout)
    if self.access_token_expired:
      raise AuthenticationError('No valid credentials provided to authorize')
    self.http = self.credentials.authorize(self.http)
    self.service = build('drive', 'v2', http=self.http)

  def Get_Http_Object(self):
    """Create and authorize an httplib2.Http object. Necessary for
    thread-safety.
    :return: The http object to be used in each call.
    :rtype: httplib2.Http
    """
    http = httplib2.Http(timeout=self.http_timeout)
    http = self.credentials.authorize(http)
    return http



from yaml import load
from yaml import YAMLError
try:
  from yaml import CLoader as Loader
except ImportError:
  from yaml import Loader

SETTINGS_FILE = 'settings.yaml'
SETTINGS_STRUCT = {
    'client_config_backend': {
        'type': str,
        'required': True,
        'default': 'file',
        'dependency': [
            {
                'value': 'file',
                'attribute': ['client_config_file']
            },
            {
                'value': 'settings',
                'attribute': ['client_config']
            },
            {
                'value': 'service',
                'attribute': ['service_config']
            }
        ]
    },
    'save_credentials': {
        'type': bool,
        'required': True,
        'default': False,
        'dependency': [
            {
                'value': True,
                'attribute': ['save_credentials_backend']
            }
        ]
    },
    'get_refresh_token': {
        'type': bool,
        'required': False,
        'default': False
    },
    'client_config_file': {
        'type': str,
        'required': False,
        'default': 'client_secrets.json'
    },
    'save_credentials_backend': {
        'type': str,
        'required': False,
        'dependency': [
            {
                'value': 'file',
                'attribute': ['save_credentials_file']
            }
        ]
    },
    'client_config': {
        'type': dict,
        'required': False,
        'struct': {
            'client_id': {
                'type': str,
                'required': True
            },
            'client_secret': {
                'type': str,
                'required': True
            },
            'auth_uri': {
                'type': str,
                'required': True,
                'default': 'https://accounts.google.com/o/oauth2/auth'
            },
            'token_uri': {
                'type': str,
                'required': True,
                'default': 'https://accounts.google.com/o/oauth2/token'
            },
            'redirect_uri': {
                'type': str,
                'required': True,
                'default': 'urn:ietf:wg:oauth:2.0:oob'
            },
            'revoke_uri': {
                'type': str,
                'required': True,
                'default': None
            }
        }
    },
    'service_config': {
        'type': dict,
        'required': False,
        'struct': {
            'client_user_email': {
                'type': str,
                'required': True,
                'default': None
            },
            'client_service_email': {
                'type': str,
                'required': True
            },
            'client_pkcs12_file_path': {
                'type': str,
                'required': True
            }
        }
    },
    'oauth_scope': {
        'type': list,
        'required': True,
        'struct': str,
        'default': ['https://www.googleapis.com/auth/drive']
    },
    'save_credentials_file': {
        'type': str,
        'required': False,
    }
}


class SettingsError(IOError):
  """Error while loading/saving settings"""


class InvalidConfigError(IOError):
  """Error trying to read client configuration."""


def LoadSettingsFile(filename=SETTINGS_FILE):
  """Loads settings file in yaml format given file name.

  :param filename: path for settings file. 'settings.yaml' by default.
  :type filename: str.
  :raises: SettingsError
  """
  try:
    stream = open(filename, 'r')
    data = load(stream, Loader=Loader)
  except (YAMLError, IOError) as e:
    raise SettingsError(e)
  return data


def ValidateSettings(data):
  """Validates if current settings is valid.

  :param data: dictionary containing all settings.
  :type data: dict.
  :raises: InvalidConfigError
  """
  _ValidateSettingsStruct(data, SETTINGS_STRUCT)


def _ValidateSettingsStruct(data, struct):
  """Validates if provided data fits provided structure.

  :param data: dictionary containing settings.
  :type data: dict.
  :param struct: dictionary containing structure information of settings.
  :type struct: dict.
  :raises: InvalidConfigError
  """
  # Validate required elements of the setting.
  for key in struct:
    if struct[key]['required']:
      _ValidateSettingsElement(data, struct, key)


def _ValidateSettingsElement(data, struct, key):
  """Validates if provided element of settings data fits provided structure.

  :param data: dictionary containing settings.
  :type data: dict.
  :param struct: dictionary containing structure information of settings.
  :type struct: dict.
  :param key: key of the settings element to validate.
  :type key: str.
  :raises: InvalidConfigError
  """
  # Check if data exists. If not, check if default value exists.
  value = data.get(key)
  data_type = struct[key]['type']
  if value is None:
    try:
      default = struct[key]['default']
    except KeyError:
      raise InvalidConfigError('Missing required setting %s' % key)
    else:
      data[key] = default
  # If data exists, Check type of the data
  elif type(value) is not data_type:
    raise InvalidConfigError('Setting %s should be type %s' % (key, data_type))
  # If type of this data is dict, check if structure of the data is valid.
  if data_type is dict:
    _ValidateSettingsStruct(data[key], struct[key]['struct'])
  # If type of this data is list, check if all values in the list is valid.
  elif data_type is list:
    for element in data[key]:
      if type(element) is not struct[key]['struct']:
        raise InvalidConfigError('Setting %s should be list of %s' %
                                 (key, struct[key]['struct']))
  # Check dependency of this attribute.
  dependencies = struct[key].get('dependency')
  if dependencies:
    for dependency in dependencies:
      if value == dependency['value']:
        for reqkey in dependency['attribute']:
          _ValidateSettingsElement(data, struct, reqkey)



import unittest

from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive

class ApiAttributeTest(unittest.TestCase):
    """Test ApiAttr functions.
    """
    ga = GoogleAuth('settings/test1.yaml')
    ga.LocalWebserverAuth()
    first_file = 'a.png'
    second_file = 'b.png'

    def test_UpdateMetadataNotInfinitelyNesting(self):
        # Verify 'metadata' field present.
        self.assertTrue(self.file1.metadata is not None)
        self.file1.UpdateMetadata()
        self.file1.UpdateMetadata()

        # Verify 'metadata' field still present.
        self.assertTrue(self.file1.metadata is not None)
        # Ensure no 'metadata' field in 'metadata' (i.e. nested).
        self.assertTrue('metadata' not in self.file1.metadata)

    def setUp(self):
        self.drive = GoogleDrive(self.ga)
        self.file1 = self.drive.CreateFile()
        self.file1.Upload()

    def tearDown(self):
        self.file1.Delete()

if __name__ == '__main__':
    unittest.main()


import random
import re

newline_pattern = re.compile(r'[\r\n]')


def CreateRandomFileName():
    hash = random.getrandbits(128)
    return "%032x" % hash


def StripNewlines(string):
    return newline_pattern.sub("", string)



import io
import mimetypes

from apiclient import errors
from apiclient.http import MediaIoBaseUpload
from functools import wraps

from .apiattr import ApiAttribute
from .apiattr import ApiAttributeMixin
from .apiattr import ApiResource
from .apiattr import ApiResourceList
from .auth import LoadAuth

BLOCK_SIZE = 1024
# Usage: MIME_TYPE_TO_BOM['<Google Drive mime type>']['<download mimetype>'].
MIME_TYPE_TO_BOM = {
  'application/vnd.google-apps.document': {
    'text/plain': u'\ufeff'.encode('utf8')
  }
}

class FileNotUploadedError(RuntimeError):
  """Error trying to access metadata of file that is not uploaded."""


class ApiRequestError(IOError):
  """Error while making any API requests."""


class FileNotDownloadableError(RuntimeError):
  """Error trying to download file that is not downloadable."""


def LoadMetadata(decoratee):
  """Decorator to check if the file has metadata and fetches it if not.

  :raises: ApiRequestError, FileNotUploadedError
  """
  @wraps(decoratee)
  def _decorated(self, *args, **kwargs):
    if not self.uploaded:
      self.FetchMetadata()
    return decoratee(self, *args, **kwargs)
  return _decorated


class GoogleDriveFileList(ApiResourceList):
  """Google Drive FileList instance.

  Equivalent to Files.list() in Drive APIs.
  """

  def __init__(self, auth=None, param=None):
    """Create an instance of GoogleDriveFileList."""
    super(GoogleDriveFileList, self).__init__(auth=auth, metadata=param)

  @LoadAuth
  def _GetList(self):
    """Overwritten method which actually makes API call to list files.

    :returns: list -- list of pydrive.files.GoogleDriveFile.
    """
    self.metadata = self.auth.service.files().list(**dict(self)).execute(
      http=self.http)
    result = []
    for file_metadata in self.metadata['items']:
      tmp_file = GoogleDriveFile(
          auth=self.auth,
          metadata=file_metadata,
          uploaded=True)
      result.append(tmp_file)
    return result


class GoogleDriveFile(ApiAttributeMixin, ApiResource):
  """Google Drive File instance.

  Inherits ApiResource which inherits dict.
  Can access and modify metadata like dictionary.
  """
  content = ApiAttribute('content')
  uploaded = ApiAttribute('uploaded')
  metadata = ApiAttribute('metadata')

  def __init__(self, auth=None, metadata=None, uploaded=False):
    """Create an instance of GoogleDriveFile.

    :param auth: authorized GoogleAuth instance.
    :type auth: pydrive.auth.GoogleAuth
    :param metadata: file resource to initialize GoogleDriveFile with.
    :type metadata: dict.
    :param uploaded: True if this file is confirmed to be uploaded.
    :type uploaded: bool.
    """
    ApiAttributeMixin.__init__(self)
    ApiResource.__init__(self)
    self.metadata = {}
    self.dirty = {'content': False}
    self.auth = auth
    self.uploaded = uploaded
    if uploaded:
      self.UpdateMetadata(metadata)
    elif metadata:
      self.update(metadata)
    self._ALL_FIELDS = 'alternateLink,appDataContents,' \
                      'canComment,canReadRevisions,' \
                 'copyable,createdDate,defaultOpenWithLink,description,' \
                 'downloadUrl,editable,embedLink,etag,explicitlyTrashed,' \
                 'exportLinks,fileExtension,fileSize,folderColorRgb,' \
                 'fullFileExtension,headRevisionId,iconLink,id,' \
                 'imageMediaMetadata,indexableText,isAppAuthorized,kind,' \
                 'labels,lastModifyingUser,lastModifyingUserName,' \
                 'lastViewedByMeDate,markedViewedByMeDate,md5Checksum,' \
                 'mimeType,modifiedByMeDate,modifiedDate,openWithLinks,' \
                 'originalFilename,ownedByMe,ownerNames,owners,parents,' \
                 'permissions,properties,quotaBytesUsed,selfLink,shareable,' \
                 'shared,sharedWithMeDate,sharingUser,spaces,thumbnail,' \
                 'thumbnailLink,title,userPermission,version,' \
                 'videoMediaMetadata,webContentLink,webViewLink,writersCanShare'
    self.has_bom = True

  def __getitem__(self, key):
    """Overwrites manner of accessing Files resource.

    If this file instance is not uploaded and id is specified,
    it will try to look for metadata with Files.get().

    :param key: key of dictionary query.
    :type key: str.
    :returns: value of Files resource
    :raises: KeyError, FileNotUploadedError
    """
    try:
      return dict.__getitem__(self, key)
    except KeyError as e:
      if self.uploaded:
        raise KeyError(e)
      if self.get('id'):
        self.FetchMetadata()
        return dict.__getitem__(self, key)
      else:
        raise FileNotUploadedError()

  def SetContentString(self, content, encoding='utf-8'):
    """Set content of this file to be a string.

    Creates io.BytesIO instance of utf-8 encoded string.
    Sets mimeType to be 'text/plain' if not specified.

    :param encoding: The encoding to use when setting the content of this file.
    :type encoding: str
    :param content: content of the file in string.
    :type content: str
    """
    self.content = io.BytesIO(content.encode(encoding))
    if self.get('mimeType') is None:
      self['mimeType'] = 'text/plain'

  def SetContentFile(self, filename):
    """Set content of this file from a file.

    Opens the file specified by this method.
    Will be read, uploaded, and closed by Upload() method.
    Sets metadata 'title' and 'mimeType' automatically if not specified.

    :param filename: name of the file to be uploaded.
    :type filename: str.
    """
    self.content = open(filename, 'rb')
    if self.get('title') is None:
      self['title'] = filename
    if self.get('mimeType') is None:
      self['mimeType'] = mimetypes.guess_type(filename)[0]

  def GetContentString(self, mimetype=None, encoding='utf-8', remove_bom=False):
    """Get content of this file as a string.

    :param mimetype: The mimetype of the content string.
    :type mimetype: str

    :param encoding: The encoding to use when decoding the byte string.
    :type encoding: str

    :param remove_bom: Whether to strip a known BOM.
    :type remove_bom: bool

    :returns: str -- utf-8 decoded content of the file
    :raises: ApiRequestError, FileNotUploadedError, FileNotDownloadableError
    """
    if self.content is None or \
                    type(self.content) is not io.BytesIO or \
                    self.has_bom == remove_bom:
      self.FetchContent(mimetype, remove_bom)
    return self.content.getvalue().decode(encoding)

  def GetContentFile(self, filename, mimetype=None, remove_bom=False):
    """Save content of this file as a local file.

    :param filename: name of the file to write to.
    :type filename: str
    :param mimetype: mimeType of the file.
    :type mimetype: str
    :param remove_bom: Whether to remove the byte order marking.
    :type remove_bom: bool
    :raises: ApiRequestError, FileNotUploadedError, FileNotDownloadableError
    """
    if self.content is None or \
                    type(self.content) is not io.BytesIO or \
                    self.has_bom == remove_bom:
      self.FetchContent(mimetype, remove_bom)
    f = open(filename, 'wb')
    f.write(self.content.getvalue())
    f.close()

  @LoadAuth
  def FetchMetadata(self, fields=None, fetch_all=False):
    """Download file's metadata from id using Files.get().

    :param fields: The fields to include, as one string, each entry separated
    by commas, e.g. 'fields,labels'.
    :type fields: str

    :param fetch_all: Whether to fetch all fields.
    :type fetch_all: bool

    :raises: ApiRequestError, FileNotUploadedError
    """
    file_id = self.metadata.get('id') or self.get('id')

    if fetch_all:
      fields = self._ALL_FIELDS

    if file_id:
      try:
        metadata = self.auth.service.files().get(fileId=file_id,
                                                 fields=fields)\
          .execute(http=self.http)
      except errors.HttpError as error:
        raise ApiRequestError(error)
      else:
        self.uploaded = True
        self.UpdateMetadata(metadata)
    else:
      raise FileNotUploadedError()

  @LoadMetadata
  def FetchContent(self, mimetype=None, remove_bom=False):
    """Download file's content from download_url.

    :raises: ApiRequestError, FileNotUploadedError, FileNotDownloadableError
    """
    download_url = self.metadata.get('downloadUrl')
    export_links = self.metadata.get('exportLinks')
    if download_url:
      self.content = io.BytesIO(self._DownloadFromUrl(download_url))
      self.dirty['content'] = False

    elif export_links and export_links.get(mimetype):
      self.content = io.BytesIO(
          self._DownloadFromUrl(export_links.get(mimetype)))
      self.dirty['content'] = False

    else:
      raise FileNotDownloadableError(
        'No downloadLink/exportLinks for mimetype found in metadata')

    if mimetype == 'text/plain' and remove_bom:
        self._RemovePrefix(self.content,
                           MIME_TYPE_TO_BOM[self['mimeType']][mimetype])
        self.has_bom = not remove_bom

  def Upload(self, param=None):
    """Upload/update file by choosing the most efficient method.

    :param param: additional parameter to upload file.
    :type param: dict.
    :raises: ApiRequestError
    """
    if self.uploaded or self.get('id') is not None:
      if self.dirty['content']:
        self._FilesUpdate(param=param)
      else:
        self._FilesPatch(param=param)
    else:
      self._FilesInsert(param=param)

  def Trash(self, param=None):
    """Move a file to the trash.

    :raises: ApiRequestError
    """
    self._FilesTrash(param=param)

  def UnTrash(self, param=None):
    """Move a file out of the trash.
    :param param: Additional parameter to file.
    :type param: dict.
    :raises: ApiRequestError
    """
    self._FilesUnTrash(param=param)

  def Delete(self, param=None):
    """Hard-delete a file.

    :param param: additional parameter to file.
    :type param: dict.
    :raises: ApiRequestError
    """
    self._FilesDelete(param=param)

  def InsertPermission(self, new_permission):
    """Insert a new permission. Re-fetches all permissions after call.

    :param new_permission: The new permission to insert, please see the
    official Google Drive API guide on permissions.insert for details.

    :type new_permission: object

    :return: The permission object.
    :rtype: object
    """
    file_id = self.metadata.get('id') or self['id']
    try:
      permission = self.auth.service.permissions().insert(
        fileId=file_id, body=new_permission).execute(http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      self.GetPermissions()  # Update permissions field.

    return permission

  @LoadAuth
  def GetPermissions(self):
    """Downloads all permissions from Google Drive, as this information is
    not downloaded by FetchMetadata by default.

    :return: A list of the permission objects.
    :rtype: object[]
    """
    self.FetchMetadata(fields='permissions')
    return self.metadata.get('permissions')

  def DeletePermission(self, permission_id):
    """Deletes the permission specified by the permission_id.

    :param permission_id: The permission id.
    :type permission_id: str
    :return: True if it succeeds.
    :rtype: bool
    """
    return self._DeletePermission(permission_id)

  @LoadAuth
  def _FilesInsert(self, param=None):
    """Upload a new file using Files.insert().

    :param param: additional parameter to upload file.
    :type param: dict.
    :raises: ApiRequestError
    """
    if param is None:
      param = {}
    param['body'] = self.GetChanges()
    try:
      if self.dirty['content']:
        param['media_body'] = self._BuildMediaBody()
      metadata = self.auth.service.files().insert(**param).execute(
        http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      self.uploaded = True
      self.dirty['content'] = False
      self.UpdateMetadata(metadata)

  @LoadAuth
  def _FilesUnTrash(self, param=None):
    """Un-delete (Trash) a file using Files.UnTrash().
    :param param: additional parameter to file.
    :type param: dict.
    :raises: ApiRequestError
    """
    if param is None:
      param = {}
    param['fileId'] = self.metadata.get('id') or self['id']
    try:
      self.auth.service.files().untrash(**param).execute(
        http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      if self.metadata:
        self.metadata[u'labels'][u'trashed'] = False
      return True

  @LoadAuth
  def _FilesTrash(self, param=None):
    """Soft-delete (Trash) a file using Files.Trash().

    :param param: additional parameter to file.
    :type param: dict.
    :raises: ApiRequestError
    """
    if param is None:
      param = {}
    param['fileId'] = self.metadata.get('id') or self['id']
    try:
      self.auth.service.files().trash(**param).execute(
        http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      if self.metadata:
        self.metadata[u'labels'][u'trashed'] = True
      return True

  @LoadAuth
  def _FilesDelete(self, param=None):
    """Delete a file using Files.Delete()
    (WARNING: deleting permanently deletes the file!)

    :param param: additional parameter to file.
    :type param: dict.
    :raises: ApiRequestError
    """
    if param is None:
      param = {}
    param['fileId'] = self.metadata.get('id') or self['id']
    try:
      self.auth.service.files().delete(**param).execute(http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      return True

  @LoadAuth
  @LoadMetadata
  def _FilesUpdate(self, param=None):
    """Update metadata and/or content using Files.Update().

    :param param: additional parameter to upload file.
    :type param: dict.
    :raises: ApiRequestError, FileNotUploadedError
    """
    if param is None:
      param = {}
    param['body'] = self.GetChanges()
    param['fileId'] = self.metadata.get('id')
    try:
      if self.dirty['content']:
        param['media_body'] = self._BuildMediaBody()
      metadata = self.auth.service.files().update(**param).execute(
        http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      self.uploaded = True
      self.dirty['content'] = False
      self.UpdateMetadata(metadata)

  @LoadAuth
  @LoadMetadata
  def _FilesPatch(self, param=None):
    """Update metadata using Files.Patch().

    :param param: additional parameter to upload file.
    :type param: dict.
    :raises: ApiRequestError, FileNotUploadedError
    """
    if param is None:
      param = {}
    param['body'] = self.GetChanges()
    param['fileId'] = self.metadata.get('id')
    try:
      metadata = self.auth.service.files().patch(**param).execute(
        http=self.http)
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      self.UpdateMetadata(metadata)

  def _BuildMediaBody(self):
    """Build MediaIoBaseUpload to get prepared to upload content of the file.

    Sets mimeType as 'application/octet-stream' if not specified.

    :returns: MediaIoBaseUpload -- instance that will be used to upload content.
    """
    if self.get('mimeType') is None:
      self['mimeType'] = 'application/octet-stream'
    return MediaIoBaseUpload(self.content, self['mimeType'], resumable=True)

  @LoadAuth
  def _DownloadFromUrl(self, url):
    """Download file from url using provided credential.

    :param url: link of the file to download.
    :type url: str.
    :returns: str -- content of downloaded file in string.
    :raises: ApiRequestError
    """
    resp, content = self.http.request(url)
    if resp.status != 200:
      raise ApiRequestError('Cannot download file: %s' % resp)
    return content

  @LoadAuth
  def _DeletePermission(self, permission_id):
    """Deletes the permission remotely, and from the file object itself.

    :param permission_id: The ID of the permission.
    :type permission_id: str

    :return: The permission
    :rtype: object
    """
    file_id = self.metadata.get('id') or self['id']
    try:
      self.auth.service.permissions().delete(
        fileId=file_id, permissionId=permission_id).execute()
    except errors.HttpError as error:
      raise ApiRequestError(error)
    else:
      if 'permissions' in self and 'permissions' in self.metadata:
        permissions = self['permissions']
        is_not_current_permission = lambda per: per['id'] == permission_id
        permissions = filter(is_not_current_permission, permissions)
        self['permissions'] = permissions
        self.metadata['permissions'] = permissions
      return True

  @staticmethod
  def _RemovePrefix(file_object, prefix, block_size=BLOCK_SIZE):
    """Deletes passed prefix by shifting content of passed file object by to
    the left. Operation is in-place.

    Args:
      file_object (obj): The file object to manipulate.
      prefix (str): The prefix to insert.
      block_size (int): The size of the blocks which are moved one at a time.
    """
    prefix_length = len(prefix)
    # Detect if prefix exists in file.
    content_start = file_object.read(prefix_length)

    if content_start == prefix:
      # Shift content left by prefix length, by copying 1KiB at a time.
      block_to_write = file_object.read(block_size)
      current_block_length = len(block_to_write)

      # Read and write location in separate variables for simplicity.
      read_location = prefix_length + current_block_length
      write_location = 0

      while current_block_length > 0:
        # Write next block.
        file_object.seek(write_location)
        file_object.write(block_to_write)
        # Set write location to the next block.
        write_location += len(block_to_write)

        # Read next block of input.
        file_object.seek(read_location)
        block_to_write = file_object.read(block_size)
        # Update the current block length and read_location.
        current_block_length = len(block_to_write)
        read_location += current_block_length

      # Truncate the file to its, now shorter, length.
      file_object.truncate(read_location - prefix_length)

  @staticmethod
  def _InsertPrefix(file_object, prefix, block_size=BLOCK_SIZE):
    """Inserts the passed prefix in the beginning of the file, operation is
    in-place.

    Args:
      file_object (obj): The file object to manipulate.
      prefix (str): The prefix to insert.
    """
    # Read the first two blocks.
    first_block = file_object.read(block_size)
    second_block = file_object.read(block_size)
    # Pointer to the first byte of the next block to be read.
    read_location = block_size * 2

    # Write BOM.
    file_object.seek(0)
    file_object.write(prefix)
    # {read|write}_location separated for readability.
    write_location = len(prefix)

    # Write and read block alternatingly.
    while len(first_block):
      # Write first block.
      file_object.seek(write_location)
      file_object.write(first_block)
      # Increment write_location.
      write_location += block_size

      # Move second block into first variable.
      first_block = second_block

      # Read in the next block.
      file_object.seek(read_location)
      second_block = file_object.read(block_size)
      # Increment read_location.
      read_location += block_size
